﻿using Buddy.Wildstar.Game;
using Buddy.Wildstar.Game.Actors;
using log4net;
using System;
using System.Linq;
using Spell = Buddy.Wildstar.Game.SpellManager.SpellInformation;

namespace Agility
{
    static class SpellController
    {
        private static Player Me
        {
            get
            {
                return GameManager.LocalPlayer;
            }
        }
        private static ILog Log = LogManager.GetLogger("Agility [SpellController]");

        public static int BuffCount(string buffName, uint durationRemaining = 0)
        {
            int count = 0;
            foreach (Buff buff in GameManager.LocalPlayer.Buffs)
            {
                if (buff.Name != buffName)
                    continue;

                if (buff.TimeLeft.TotalMilliseconds < durationRemaining)
                    continue;

                count += (int)buff.StackCount;
            }
            return count;
        }

        public static bool CanCast(string spellName, bool instant = false)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return false;

            if (spell.CooldownFlags == CastResult.Ok) // OK
                return true;

            return false;
        }

        public static void Cast(string spellName)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return;

            Log.Info(String.Format("[{0}] [{1}] [{2}]\t{3}",
                spell.TierId,
                spell.Tier,
                spell.CooldownFlags,
                spell.Name
            ));

            spell.Cast();
        }

        public static int Cooldown(string spellName)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return 10000;

            return spell.CooldownRemaining;
        }

        public static bool HasAMP(string AugName)
        {
            // Credit : https://www.thebuddyforum.com/members/337126-lukov.html
            return GameManager.Lua.GetReturnValue<bool>(
              @"for _,AMP in pairs(AbilityBook.GetEldanAugmentationData(AbilityBook.GetCurrentSpec()).tAugments) do
                    if (AMP.strTitle == '" + AugName + @"') then
                        return (AMP.eEldanAvailability == AbilityBook.CodeEnumEldanAvailability.Activated)
                    end
                end
                return false
            ");
        }

        public static bool HasSpell(string spellName)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return false;

            return spell.Tier > 0;
        }
        
        public static bool IsFullyCharged(string spellName)
        {
            // Ensure we are casting and the name matches.
            if (Me.CurrentCast == null || Me.CurrentCast.Name != spellName)
                return false;

            int CurrentCastElapsed = (int)GameManager.Lua.GetReturnValue<float>("return GameLib.GetPlayerUnit():GetCastElapsed()");
            int CastThreshold = (int)GameManager.Lua.GetReturnValue<float>(string.Format("return GameLib.GetSpell({0}):GetThresholdTime()*1000",
                Me.CurrentCast.SpellId
            ));

            return CurrentCastElapsed >= CastThreshold;
        }

        public static bool IsTicking(string buffName, int durationLeft = 0)
        {
            return GameManager.Actors.Values
                .Any(a => a.Buffs
                    .Any(b => b.Name == buffName && b.Caster.Guid == Me.Guid && b.TimeLeft.TotalMilliseconds > durationLeft));
        }

        public static bool IsTicking(int spellId, int durationLeft = 0)
        {
            return GameManager.Actors.Values
                .Any(a => a.Buffs
                    .Any(b => b.SpellId == spellId && b.Caster.Guid == Me.Guid && b.TimeLeft.TotalMilliseconds > durationLeft));
        }

        public static void Release(string spellName)
        {
            if (!GameManager.LocalPlayer.IsCasting)
                return;

            SpellManager.GetSpell(spellName).Cast(false);
        }

        public static void SetStance(string spell)
        {
            GameManager.Lua.DoString(
              @"for idx, spellObject in pairs(GameLib.GetClassInnateAbilitySpells().tSpells) do
                    if idx % 2 == 0 and spellObject:GetName() == '"+spell+@"' then
                        GameLib.SetCurrentClassInnateAbilityIndex(idx/2)
                    end
                end
            ");
        }

        public static int Tier(string spellName)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return 0;

            return spell.Tier;
        }

        public static string GetGadgetName()
        {
            Spell gadgetSpell = SpellManager.GetMiscSpell(MiscSpellType.Gadget);
            if (gadgetSpell == null || !gadgetSpell.IsValid)
                return null;

            return gadgetSpell.Name;
        }

		public static double LastCastMS(string spellName)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return 0;

            return DateTime.UtcNow.Subtract(spell.LastCast).TotalMilliseconds;
        }

        public static bool Missed(string spellName)
        {
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                return false;

            return spell.Missed;
        }


    }
}
