﻿using Buddy.Common.Math;
using Buddy.Wildstar.Engine;
using Buddy.Wildstar.Game;
using Buddy.Wildstar.Game.Actors;
using log4net;
using System;
using System.Collections.Generic;
using System.Linq;
using Agility.Events;

namespace Agility.Routines
{
    class Spellslinger_fDPS : IAgilityRoutine
    {
        private static ILog Log = LogManager.GetLogger("Agility [SpellController]");

        public string Name { get { return "Sling Ding Damage"; } }
        public string Version { get { return "0.1"; } }
        public string Author { get { return "flawd"; } }
		public Class Class { get { return Class.Spellslinger; } }      

        public static class RoutineSettings
        {
            // Debug Settings
            public static bool Debug = true;
            public static bool DebugSurge = false; //1 
            public static bool DebugIgnite = true; //2
            public static bool DebugCasts = false; //3
            public static bool DebugRoutine = true; //4
            public static int SpamBuffer = 1000;

            //Rotation Settings
            public static bool PullChargedShot = false;         // If out of combat pull with Charged shot.

            public static bool ReserveForIgnite = true;         // Try to save SpellPower to always be able to cast Ignite  
            public static int IgniteDurationDelay = 0;          // Tune window for ignite debuff duration
            public static int IgniteHitBuffer = 300;            // Max Delay for Ignite Debuff to appear on Target.
            public static int IgniteReserveSP = 19;
            public static int IgniteReserveDuration = 2000;

            public static int SurgeBuffer = 300;                // Max MS to wait for surge checks   

            //Assault Spell Settings
            public static bool QuickDrawSurge = true;           // Spend Spell Power on Skill?
            public static int QuickDrawSP = 45;                 // Min Spell Power for cast.
            public static bool ChargedShotSurge = true;
            public static bool ChargedShotOnlySurge = true;     // Prevent Casting Spell while un-surged.
            public static int ChargedShotSP = 25;
            public static bool IgniteSurge = true;
            public static int IgniteSP = 25;            
            public static bool WildBarrageSurge = false;
            public static bool WildBarrageOnlySurge = false;
            public static int WildBarrageSP = 35;
            public static bool ChillSurge = false;
            public static bool ChillOnlySurge = false;
            public static int ChillSP = 40;
            public static bool RapidFireSurge = false;
            public static bool RapidFireOnlySurge = false;
            public static int RapidFireSP = 35;
            public static bool TrueShotSurge = false;
            public static bool TrueShotOnlySurge = false;
            public static int TrueShotSP = 35;
            public static bool ArcaneMissilesSurge = false;
            public static bool ArcaneMissilesOnlySurge = false;
            public static int ArcaneMissilesSP = 35;
            public static bool AssassinateSurge = true;
            public static int AssassinateSurgeSP = 25;
        }

        private Player Me { get { return GameManager.LocalPlayer; } }
        private Actor Target;

        private static DateTime LastSentMSG = DateTime.Now;
        private static DateTime SurgeWait = DateTime.Now;
        private static int LastMSGSurge = -1;
        private static int LastMSGIgnite = -1;
        private static int LastMSGCast = -1;
        private static int LastMSGRoutine = -1;
        private static int TargetSurge = 0;

        private bool Initilized = false;

        private void Initilize()
        {
            if (Initilized == true)
                return;


            AgilityEvent temp = null;
            temp = new AgilityEvent(LuaEvent.StartSpellThreshold, HandleThresholdStart);
            AgilityEvents.RegisterEvent(temp);
            temp = new AgilityEvent(LuaEvent.UpdateSpellThreshold, HandleThresholdUpdate);
            AgilityEvents.RegisterEvent(temp);
            temp = new AgilityEvent(LuaEvent.ClearSpellThreshold, HandleThresholdClear);
            AgilityEvents.RegisterEvent(temp);
            temp = new AgilityEvent(LuaEvent.AttackMissed, HandleAttackMissed);
            AgilityEvents.RegisterEvent(temp);
            Initilized = true;
            rLog("Initilized Events", 0, 0);
        }

        public void rLog(string msg, int ID, int mID)
        {
            int lastMSG = -1;
            bool log = false;
            if (!RoutineSettings.Debug)
                return;

            switch (ID)
            {
                case 0:
                    log = true;            
                    break;
                case 1:
                    log = RoutineSettings.DebugSurge;
                    lastMSG = LastMSGSurge;
                    break;
                case 2:
                    log = RoutineSettings.DebugIgnite;
                    lastMSG = LastMSGIgnite;
                    break;
                case 3:
                    log = RoutineSettings.DebugCasts;
                    lastMSG = LastMSGCast;
                    break;
                case 4:
                    log = RoutineSettings.DebugRoutine;
                    lastMSG = LastMSGRoutine;
                    break;
            }

            if (log && (lastMSG != mID || DateTime.Now.Subtract(LastSentMSG).TotalMilliseconds > RoutineSettings.SpamBuffer || mID == 0))
            {
                Log.Info(string.Format(">> [{0:HH:mm:ss.f}] [{1}|{2:00}] >> {3} <<", DateTime.Now, ID, mID, msg));
                LastSentMSG = DateTime.Now;
                if (mID != 0)
                {
                    switch (ID)
                    {
                        case 1:
                            LastMSGSurge = mID;
                            break;
                        case 2:
                            LastMSGIgnite = mID;
                            break;
                        case 3:
                            LastMSGCast = mID;
                            break;
                        case 4:
                            LastMSGRoutine = mID;
                            break;
                    }
                }
            }
        }

        public string Routine()
        {

            if (Initilized != true)
                Initilize();

            Target = GameManager.LocalPlayer.CurrentTarget;
            int SPMod = 0;
            if (Target == null)
                Target = Me;

            if (RoutineSettings.IgniteSurge == true && RoutineSettings.ReserveForIgnite == true && !IgniteDue(1) && SpellController.LastCastMS("Ignite") > (8000 - RoutineSettings.IgniteReserveDuration))
            {
                SPMod = RoutineSettings.IgniteReserveSP;
            }

            // Log current Spell Surge and Spell Power state
            string msg = string.Format("Surged: [{0}] SpellPower: [{1:000}] ", Me.HasBuff("Spell Surge"), (int)Me.SpellPower);


            // If player can Flame Burst
            if (SpellController.CanCast("Flame Burst"))
            {
                rLog(msg + ">> Flame Burst!", 4, 1);
                return "Flame Burst";
            }

            // If player is currently casting
            if (Me.IsCasting)
            {
                msg = msg + string.Format(">> Casting [{0}] >> Thresshold: [{1}] ", Me.CurrentCast.Name, CastBar.CurrentThresshold);

                // Charge Shot: charge up to 3
                if (Me.CurrentCast.Name == "Charged Shot")
                {
                    if (CastBar.CurrentThresshold != 3)
                    {
                        rLog(msg + ">> Waiting for charge up!", 3, CastBar.CurrentThresshold + 2);
                        return null;
                    }
                    else
                    {
                        rLog(string.Format("{0}>> Fire! CT: [{1}]", msg, CastBar.Elapsed), 3, 0);
                        return "Charged Shot";
                    }
                }

                // True Shot: Spam it! Maybe some timing later?
                if (Me.CurrentCast.Name == "True Shot" && SpellController.CanCast("True Shot"))
                {
                    rLog(msg + ">> Fire!", 3, 0);
                    return "True Shot";
                }

                // True Shot: Spam it! Maybe some timing later?
                if (Me.CurrentCast.Name == "Rapid Fire" && SpellController.CanCast("Rapid Fire"))
                {
                    rLog(msg + ">> Fire!", 3, 0);
                    return "Rapid Fire";
                }

                // wait for current cast to mostly finish
                if (CastBar.CastRemaining > RoutineSettings.SurgeBuffer)
                {
                    rLog(string.Format("{0}>> Wait for Cast! T-{1:0.00}", msg, CastBar.CastRemaining / 1000.0), 3, CastBar.CastRemaining / 100);
                    return null;
                }
            }

            if (!SafeToSurge())
            {
                rLog(msg + ">> Wait for Surge Buffer!", 4, 1);
                return null;
            }

            // Pull with Charged Shot
            if (RoutineSettings.PullChargedShot == true && SpellController.CanCast("Charged Shot") && !Me.IsInCombat)
            {
                rLog(msg + ">> Pull with charged shot", 4, 0);
                return "Charged Shot";
            }

            if (SpellController.CanCast(SpellController.GetGadgetName()))
            {
                rLog(msg + ">> Use Gadget (Default)", 4, 0);
                return SpellController.GetGadgetName();
            }

            // If currently Surged
            if (Me.HasBuff("Spell Surge"))
            {
                msg = msg + ">> Spell Surged: ";

                if (RoutineSettings.IgniteSurge == true && SpellController.CanCast("Ignite") && Me.SpellPower >= RoutineSettings.IgniteSP && IgniteDue(1))
                {
                    rLog(msg + ">> Ignite!", 4, 0);
                    return "Ignite";
                }

                // IF Surged and need non surged Ignite drop

                if (SpellController.CanCast("Ignite") && IgniteDue(0))
                {
                    rLog(msg + ">> Unsurge for normal Ignite!", 4, 0);
                    TargetSurge = 0;
                    return "Spell Surge";
                }


                if (RoutineSettings.TrueShotSurge == true && SpellController.CanCast("True Shot") && (Me.SpellPower - SPMod) >= RoutineSettings.TrueShotSP) //18
                {
                    rLog(msg + ">> True Shot", 4, 0);
                    return "True Shot";
                }

                if (RoutineSettings.ArcaneMissilesSurge == true && SpellController.CanCast("Arcane Missiles") && (Me.SpellPower - SPMod) >= RoutineSettings.ArcaneMissilesSP) //18
                {
                    rLog(msg + ">> Arcane Missiles", 4, 0);
                    return "Arcane Missiles";
                }

                if (RoutineSettings.WildBarrageSurge == true && SpellController.CanCast("Wild Barrage") && (Me.SpellPower - SPMod) >= RoutineSettings.WildBarrageSP) //12
                {
                    rLog(msg + ">> Wild Barrage!", 4, 0);
                    return "Wild Barrage";
                }

                if (RoutineSettings.ChargedShotSurge == true && SpellController.CanCast("Charged Shot") && (Me.SpellPower - SPMod) >= RoutineSettings.ChargedShotSP) //10
                {
                    rLog(msg + ">> Charged Shot!", 4, 0);
                    return "Charged Shot";
                }

                if (RoutineSettings.AssassinateSurge == true && SpellController.CanCast("Assassinate") && (Me.SpellPower - SPMod) >= RoutineSettings.AssassinateSurgeSP && Target.HealthPercent > 30) //10
                {
                    rLog(msg + ">> Assassinate!", 4, 0);
                    return "Assassinate";
                }

                if (RoutineSettings.RapidFireSurge == true && SpellController.CanCast("Rapid Fire") && (Me.SpellPower - SPMod) >= RoutineSettings.RapidFireSP) //10
                {
                    rLog(msg + ">> Rapid Fire!", 4, 0);
                    return "Rapid Fire";
                }

                if (RoutineSettings.ChillSurge == true && SpellController.CanCast("Chill") && (Me.SpellPower - SPMod) >= RoutineSettings.ChillSP) // 7
                {
                    rLog(msg + ">> Chill!", 4, 0);
                    return "Chill";
                }

                if (RoutineSettings.QuickDrawSurge == true && SpellController.CanCast("Quick Draw") && (Me.SpellPower - SPMod) >= RoutineSettings.QuickDrawSP)
                {
                    rLog(msg + ">> Quick Draw!", 4, 0);
                    return "Quick Draw";
                }
                if (SpellController.CanCast("Spell Surge"))
                {
                    rLog(msg + ">> Unsurge!", 4, 0);
                    TargetSurge = 0;
                    return "Spell Surge";
                }
            }
            else
            {
                msg = msg + ">> Not Surged: ";

                if (SpellController.CanCast("Ignite") && IgniteDue(0))
                {
                    rLog(msg + ">> Normal Ignite!", 4, 0);
                    return "Ignite";
                }

                if (SpellController.CanCast("Assassinate") && Target.HealthPercent <= 30) //10
                {
                    rLog(msg + ">> Assassinate!", 4, 0);
                    return "Assassinate";
                }

                if (RoutineSettings.IgniteSurge == true && SpellController.CanCast("Ignite") && Me.SpellPower >= RoutineSettings.IgniteSP && IgniteDue(1) && SpellController.CanCast("Spell Surge"))
                {
                    rLog(msg + ">> Surge for Ignite!", 4, 0);
                    TargetSurge = 1;
                    return "Spell Surge";
                }

                if (SpellController.CanCast("Spell Surge") &&
                    (
                    (RoutineSettings.QuickDrawSurge == true && SpellController.CanCast("Quick Draw") && (Me.SpellPower - SPMod) >= RoutineSettings.QuickDrawSP) ||
                    (RoutineSettings.ChargedShotSurge == true && SpellController.CanCast("Charged Shot") && (Me.SpellPower - SPMod) >= RoutineSettings.ChargedShotSP) ||
                    (RoutineSettings.IgniteSurge == true && SpellController.CanCast("Ignite") && (Me.SpellPower - SPMod) >= RoutineSettings.IgniteSP && IgniteDue(1)) ||
                    (RoutineSettings.WildBarrageSurge == true && SpellController.CanCast("Wild Barrage") && (Me.SpellPower - SPMod) >= RoutineSettings.WildBarrageSP) ||
                    (RoutineSettings.ChillSurge == true && SpellController.CanCast("Chill") && (Me.SpellPower - SPMod) >= RoutineSettings.ChillSP) ||
                    (RoutineSettings.RapidFireSurge == true && SpellController.CanCast("Rapid Fire") && (Me.SpellPower - SPMod) >= RoutineSettings.RapidFireSP) ||
                    (RoutineSettings.AssassinateSurge == true && SpellController.CanCast("Assassinate") && (Me.SpellPower - SPMod) >= RoutineSettings.AssassinateSurgeSP && Target.HealthPercent > 30) ||
                    (RoutineSettings.TrueShotSurge == true && SpellController.CanCast("True Shot") && (Me.SpellPower - SPMod) >= RoutineSettings.TrueShotSP) ||
                    (RoutineSettings.ArcaneMissilesSurge == true && SpellController.CanCast("Arcane Missiles") && (Me.SpellPower - SPMod) >= RoutineSettings.ArcaneMissilesSP))
                    )
                {
                    rLog(msg + ">> Surge!", 4, 0);
                    TargetSurge = 1;
                    return "Spell Surge";
                }

                if (SpellController.CanCast("True Shot") && !RoutineSettings.TrueShotOnlySurge) //18
                {
                    rLog(msg + ">> True Shot", 4, 0);
                    return "True Shot";
                }

                if (SpellController.CanCast("Arcane Missiles") && !RoutineSettings.ArcaneMissilesOnlySurge) //18
                {
                    rLog(msg + ">> Arcane Missiles", 4, 0);
                    return "Arcane Missiles";
                }

                if (SpellController.CanCast("Wild Barrage") && !RoutineSettings.WildBarrageOnlySurge) //12
                {
                    rLog(msg + ">> Wild Barrage!", 4, 0);
                    return "Wild Barrage";
                }

                if (SpellController.CanCast("Charged Shot") && !RoutineSettings.ChargedShotOnlySurge) //10
                {
                    rLog(msg + ">> Charged Shot!", 4, 0);
                    return "Charged Shot";
                }

                if (SpellController.CanCast("Rapid Fire") && !RoutineSettings.RapidFireOnlySurge) //10
                {
                    rLog(msg + ">> Rapid Fire!", 4, 0);
                    return "Rapid Fire";
                }

                if (SpellController.CanCast("Chill") && !RoutineSettings.ChillOnlySurge) // 7  && Targeting.CombatTargets.Count > 2
                {
                    rLog(msg + ">> Chill!", 4, 0);
                    return "Chill";
                }

                if (SpellController.CanCast("Quick Draw") && !Me.HasBuff("SpellSurge"))
                {
                    rLog(msg + ">> Quick Draw!", 4, 0);
                    return "Quick Draw";
                }
            }
            return null;
        }

        // Don't normal ignite if surged is needed!
        // Check if we Need an ignite. Both normal and surged 
        private bool IgniteDue(int Surged)
        {
            int SpellTier = SpellController.Tier("Ignite");
            int SpellID = 0;
            int CastTime = 0;
            double SinceLastCast = SpellController.LastCastMS("Ignite");
            bool Missed = SpellController.Missed("Ignite");
            bool Ticking = false;

            //           0                                                       8
            //normal = 34754, 49142, 49143, 49144, 49145, 49146, 49147, 49148, 49149
            //surged = 34755, 49151, 49152, 49153, 49154, 49155, 49156, 49157, 77138

            // Get BuffID of Target Ignite Tier
            if (Surged == 1) // Surged Ignite
            {
                if (SpellTier == 8)
                    SpellID = 77138;
                else if (SpellTier == 7)
                    SpellID = 49157;
                else if (SpellTier == 6)
                    SpellID = 49156;
                else if (SpellTier == 5)
                    SpellID = 49155;
                else if (SpellTier == 4)
                    SpellID = 49154;
                else if (SpellTier == 3)
                    SpellID = 49153;
                else if (SpellTier == 2)
                    SpellID = 49152;
                else if (SpellTier == 1)
                    SpellID = 49151;
                else if (SpellTier == 0)
                    SpellID = 34755;
            }

            else if (Surged == 0) // Normal Ignite
            {
                if (SpellTier == 8)
                    return false; // Tier 8 Applies both Surged and non Surged ignore
                else if (SpellTier == 7)
                    SpellID = 49148;
                else if (SpellTier == 6)
                    SpellID = 49147;
                else if (SpellTier == 5)
                    SpellID = 49146;
                else if (SpellTier == 4)
                    SpellID = 49145;
                else if (SpellTier <= 3)
                    return false; // Can Only apply one Ignite prioritize Surged.
            }

            Ticking = SpellController.IsTicking(SpellID);

            // Pad cast time for hit recognition
            if (SpellTier >= 4)
                CastTime = 500 + RoutineSettings.IgniteHitBuffer;
            else
                CastTime = 1000 + RoutineSettings.IgniteHitBuffer;

            // Don't recast till CastTime + 250 ms has passed to prevent Spam. Only cast if target is missing Ignite or it has been 8 sec since you last cast Ignite or Ignite Missed.
            if (Missed || (SinceLastCast > CastTime && (SinceLastCast > (8000 + RoutineSettings.IgniteDurationDelay) || !Ticking)))
            {
                if (Surged == 1)
                    rLog(string.Format("IgniteDue >> Need surged Ignite!  >> Last Cast:[{0:0000}] IsTicking: [{1}] Missed: [{2}].", SinceLastCast, Ticking, Missed), 2, 2);
                else if (Surged == 0)
                    rLog(string.Format("IgniteDue >> Need normal Ignite! >> Last Cast:[{0:0000}] IsTicking:[{1}] Missed: [{2}].", SinceLastCast, Ticking, Missed), 2, 3);
                return true;
            }
            if (Surged == 1)
                rLog(string.Format("IgniteDue >> Surged Ignite Not Needed! >> Last Cast:[{0:0000}] IsTicking:[{1}] Missed: [{2}].", SinceLastCast, Ticking, Missed), 2, 1);
            else if (Surged == 0)
                rLog(string.Format("IgniteDue >> Normal Ignite Not Needed! >> Last Cast:[{0:0000}] IsTicking:[{1}] Missed: [{2}].", SinceLastCast, Ticking, Missed), 2, 1);

            return false;
        }

        private bool SafeToSurge()
        {
            int Surged;
            if (Me.HasBuff("Spell Surge"))
                Surged = 1;
            else
                Surged = 0;
            if (SpellController.LastCastMS("Spell Surge") > RoutineSettings.SurgeBuffer || TargetSurge == Surged)
            {
                rLog(string.Format("SafeToSurge >> Safe! >> Last Cast: [{0:0000}], Can Cast : [{1}]", SpellController.LastCastMS("Spell Surge"), SpellController.CanCast("Spell Surge")), 1, 1);
                return true;
            }
            rLog(string.Format("SafeToSurge >> Not Safe! >> Last Cast: [{0:0000}], Can Cast : [{1}]", SpellController.LastCastMS("Spell Surge"), SpellController.CanCast("Spell Surge")), 1, 2);
            return false;
        }


        // Threshold Handlers used for tracking casts
        private void HandleThresholdStart(string eventname, LuaEventArgument[] args) // ARGS: Charged Shot: [0:SpellID] [1:MaxCharge] [2:????] 
        {
            string str = string.Format("<< [{0:HH:mm:ss.f}] Starting Threshold <<", DateTime.Now);
            for (int i = 0; i < args.Length; i++)
                str = str + string.Format(" [{0}: {1}] ", i, args[i]);
            Log.Info(str);
            CastBar.ThresholdStart(args);
        }
        private void HandleThresholdUpdate(string eventname, LuaEventArgument[] args) // ARGS: Charged Shot: [0:SpellID] [1:CurrentCharge]
                {
            string str = string.Format("<< [{0:HH:mm:ss.f}] Updating Threshold <<", DateTime.Now);
            for (int i = 0; i < args.Length; i++)
                str = str + string.Format(" [{0}: {1}] ", i, args[i]);
            Log.Info(str);
            CastBar.ThresholdUpdate(args);
                    }
        private void HandleThresholdClear(string eventname, LuaEventArgument[] args) // ARGS: Charged Shot: [0:SpellID]
                    {
            string str = string.Format("<< [{0:HH:mm:ss.f}] Clearing Threshold <<", DateTime.Now);
            for (int i = 0; i < args.Length; i++)
                str = str + string.Format(" [{0}: {1}] ", i, args[i]);
            Log.Info(str);
            CastBar.ThresholdClear(args);
                    }

        private void HandleAttackMissed(string eventname, LuaEventArgument[] args) // ARGS  [0:Player]  [1:Target]  [2:????]  [3:SkillName] 
                    {
            string str = string.Format("<< [{0:HH:mm:ss.f}] Attack Missed <<", DateTime.Now);
            for (int i = 0; i < args.Length; i++)
                str = str + string.Format(" [{0}: {1}]", i, args[i]);
            Log.Info(str);

            string spellName = args[3];
            AgilitySpell spell = SpellCache.GetSpell(spellName);
            if (!spell.IsValid)
                spell.SetMissed();
                    }

        public static class CastBar
                    {
            public static DateTime CastStartTime;
            public static int SpellID;
            public static string SpellName;
            public static int CurrentThresshold;
            public static int MaxThresshold;
            public static int Elapsed
                    {
                get
                    {
                    if (GameManager.LocalPlayer.IsCasting)
                        return (int)GameManager.Lua.GetReturnValue<float>("return GameLib.GetPlayerUnit():GetCastElapsed()");
                    return 0;
                    }
                    }

            public static int Duration
            {
                get
                {
                    if (GameManager.LocalPlayer.IsCasting)
                        return (int)GameManager.Lua.GetReturnValue<float>("return GameLib.GetPlayerUnit():GetCastDuration()");
                    return 0;
                }
            }

            public static int CastRemaining
            {
                get
                {
                    if (GameManager.LocalPlayer.IsCasting)
                        return Duration - Elapsed;

                    return 0;
                }
            }

            public static void ThresholdStart(LuaEventArgument[] args)
            {

                if (SpellID != args[0])
                {
                    CastStartTime = DateTime.Now;
                    SpellID = args[0];
                    if (GameManager.LocalPlayer.CurrentCast != null)                    //Spell name issue with TrueShot: starts after spell cast
                        SpellName = GameManager.LocalPlayer.CurrentCast.Name;           //Pull Spell name from player. Spell IDs changed based on surge state.
                    else
                        SpellName = "???";
                    CurrentThresshold = 1;
                    MaxThresshold = args[1];

                }
            }

            public static void ThresholdUpdate(LuaEventArgument[] args)
            {
                if (SpellID == args[0])
                    CurrentThresshold = args[1];
            }

            public static void ThresholdClear(LuaEventArgument[] args)
            {
                if (SpellID == args[0])
                {
                    CastStartTime = DateTime.MinValue;
                    SpellID = 0;
                    SpellName = "";
                    CurrentThresshold = 0;
                    MaxThresshold = 0;
                }
            }
        }
    }
}

