﻿using Buddy.Wildstar.Game;
using System;
using Spell = Buddy.Wildstar.Game.SpellManager.SpellInformation;

namespace Agility
{
    class AgilitySpell
    {
        private Spell spell;
        private bool isValid;
        private string name;
        private int tier;
        private int baseId;
        private int tierId;
        //private int cooldownRemaining
        //private int cooldownFlag;
        private CastMethod castMethod;
        private DateTime lastCast;
        private float maxRange;
        private bool missed = false;

        public AgilitySpell(string spellName)
        {
            this.spell = SpellManager.GetSpell(spellName);
            this.name = spellName;

            this.isValid = (this.spell != null && this.spell.IsValid);
            if (!this.isValid)
                return;

            this.tier = spell.Tier;
            this.baseId = spell.Ability.SpellId;
            this.tierId = tier == 0 ? this.baseId : GameManager.Lua.GetReturnValue<int>("return AbilityBook.GetAbilityInfo(" + this.baseId + ").tTiers[" + (this.tier + 1) + "].splObject:GetId()");
            this.castMethod = this.spell.SpellCastMethod;
            this.lastCast = DateTime.UtcNow;
            this.maxRange = spell.TargetMaxRange;
            //this.cooldownRemaining;
            //this.cooldownFlag;
        }

        public int BaseId
        {
            get
            {
                return this.baseId;
            }
        }
        public int CastPercentDone
        {
            get
            {
                return (int)this.spell.CastPercentDone;
            }
        }
        public CastResult CooldownFlags
        {
            get
            {
                return this.spell.CooldownFlags;
            }
        }
        public int CooldownRemaining
        {
            get
            {
                return (int)GameManager.Lua.GetReturnValue<float>("return GameLib.GetSpell(" + this.tierId + "):GetCooldownRemaining()*1000");
            }
        }
        public bool IsValid
        {
            get
            {
                return this.isValid;
            }
        }
        public float MaxRange
        {
            get
            {
                return this.maxRange;
            }
        }
        public string Name
        {
            get
            {
                return this.name;
            }
        }
        public int Tier
        {
            get
            {
                return this.tier;
            }
        }
        public int TierId
        {
            get
            {
                return this.tierId;
            }
        }
        public DateTime LastCast
        {
            get
            {
                return this.lastCast;
            }
        }

        public bool Missed
        {
            get
            {
                return this.missed;
            }
        }

        public void SetMissed()
        {
            this.missed = true;
        }


        public void Cast()
        {
            switch (this.castMethod)
            {
                case CastMethod.ChargeRelease:
                    if (DateTime.UtcNow.Subtract(this.lastCast).TotalMilliseconds > 500)
                    {
                        this.lastCast = DateTime.UtcNow;
                        this.spell.Cast();
                    }
                    break;

                default:
                    this.lastCast = DateTime.UtcNow;
                    this.spell.Cast();
                    break;
            }
        }

    }
}
